//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Coord/ICoordSystem.h
//! @brief     Defines interface ICoordSystem and possible axis units.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifdef SWIG
#error no need to expose this header to Swig
#endif // SWIG
#ifndef BORNAGAIN_DEVICE_COORD_ICOORDSYSTEM_H
#define BORNAGAIN_DEVICE_COORD_ICOORDSYSTEM_H

#include "Base/Types/ICloneable.h"
#include "Device/Coord/Tags.h"
#include <map>
#include <memory>
#include <string>
#include <vector>

class IAxis;

//! Interface to provide axis translations to different units for simulation output

//! Child classes are currently declared in
//! - Device/Detector/CoordSystem2D.h,
//! - Core/Scan/CoordSystem1D.h.

class ICoordSystem : public ICloneable {
public:
    ICoordSystem(std::vector<const IAxis*> axes)
        : m_axes(axes)
    {
    }
    ~ICoordSystem() override;

    ICoordSystem* clone() const override = 0;

    size_t rank() const { return m_axes.size(); }

    virtual double calculateMin(size_t i_axis, Coords units) const = 0;
    virtual double calculateMax(size_t i_axis, Coords units) const = 0;

    virtual std::string nameOfAxis(size_t i_axis, Coords units) const = 0;

    virtual std::vector<Coords> availableUnits() const = 0;
    virtual Coords defaultUnits() const = 0;

    virtual IAxis* createConvertedAxis(size_t i_axis, Coords units) const = 0;

    std::vector<const IAxis*> defaultAxes() const;
    std::vector<const IAxis*> convertedAxes(Coords units) const;

protected:
    Coords substituteDefaultUnits(Coords units) const;

    std::vector<const IAxis*> m_axes;
};

#endif // BORNAGAIN_DEVICE_COORD_ICOORDSYSTEM_H
