//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/IO/ReadReflectometry.cpp
//! @brief     Implements class ReadWriteReflectometry.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Device/IO/ReadReflectometry.h"
#include "Base/Axis/DiscreteAxis.h"
#include "Base/Util/StringUtils.h"
#include "Device/Data/Datafield.h"
#include "Device/IO/DataFormatUtils.h"
#include <map>

// #bamigration +++ this works only if separator is space or tab; it does not
// work e.g. with comma or semicolon
Datafield* ReadReflectometry::readDatafield(std::istream& inStream)
{
    std::string line;
    std::vector<std::vector<double>> vecVec;
    std::map<double, double> QvsR;
    std::map<double, double> QvsDR;
    std::map<double, double> QvsDQ;

    // Read numbers from file:
    while (std::getline(inStream, line)) {
        line = BaseUtils::String::trim(line);
        try {
            // #bamigration +++ this works only if separator is space or tab; it does not
            // work e.g. with comma or semicolon
            std::vector<double> rowVec = DataUtils::Format::parse_doubles(line);

            // take only non-negative Q values
            if (rowVec[0] >= 0.0)
                vecVec.push_back(rowVec);
        } catch (...) { // #bamigration +++ This eats useful errors away...
            continue;
        }
    }

    // validate - There is at least one row and at least two columns
    size_t nrows = vecVec.size();
    if (nrows < 1)
        throw std::runtime_error("No numerical values found");
    size_t ncols = vecVec[0].size();
    if (ncols < 2)
        throw std::runtime_error("Minimum 2 columns required");

    // Assign Q vs R, dR, dQ:
    for (size_t row = 0; row < nrows; row++) {
        if (vecVec[row].size() != ncols)
            throw std::runtime_error("The number of columns varies among the rows");
        double Q = vecVec[row][0];
        switch (ncols) {
        case 1:
            break;
        case 2:
            QvsR[Q] = vecVec[row][1];
            QvsDR[Q] = 0;
            QvsDQ[Q] = 0;
            break;
        case 3:
            QvsR[Q] = vecVec[row][1];
            QvsDR[Q] = vecVec[row][2];
            QvsDQ[Q] = 0;
            break;
        default:
            QvsR[Q] = vecVec[row][1];
            QvsDR[Q] = vecVec[row][2];
            QvsDQ[Q] = vecVec[row][3];
            break;
        }
    }

    std::vector<double> qVec;
    std::vector<double> rVec;
    for (auto it = QvsR.begin(); it != QvsR.end(); ++it) {
        if (it->second <= 0)
            continue;
        qVec.push_back(it->first);
        rVec.push_back(it->second);
    }
    std::vector<double> eVec;
    for (auto it = QvsDR.begin(); it != QvsDR.end(); ++it) {
        if (it->second <= 0)
            continue;
        eVec.push_back(it->second);
    }
    return new Datafield{{new DiscreteAxis("qVector", qVec)}, rVec, eVec};
}
