#include "Base/Axis/FixedBinAxis.h"
#include "Base/Axis/Bin.h"
#include "Device/IO/DataFormatUtils.h"
#include "Tests/GTestWrapper/google_test.h"
#include <iostream>

TEST(FixedBinAxisTest, IndexedAccessor)
{
    FixedBinAxis a1("length", 100, 0.0, 10.0);
    EXPECT_EQ(100u, a1.size());
    EXPECT_EQ(0.0, a1.min());
    EXPECT_EQ(10.0, a1.max());
    EXPECT_DOUBLE_EQ(0.05, a1.binCenter(0));
    EXPECT_DOUBLE_EQ(0.15, a1.binCenter(1));
    EXPECT_DOUBLE_EQ(6.55, a1.binCenter(65));
    EXPECT_DOUBLE_EQ(9.95, a1.binCenter(99));

    FixedBinAxis a2("name", 3, -1.5, 1.5);
    EXPECT_DOUBLE_EQ(-1.0, a2.binCenter(0));
    EXPECT_DOUBLE_EQ(0.0, a2.binCenter(1));
    EXPECT_DOUBLE_EQ(1.0, a2.binCenter(2));
    EXPECT_THROW(a2.binCenter(3), std::runtime_error);
}

TEST(FixedBinAxisTest, VectorOfUnitLength)
{
    FixedBinAxis axis("name", 1, 1.0, 2.0);
    EXPECT_EQ(1u, axis.size());
    EXPECT_EQ(double(1.0), axis.min());
    EXPECT_EQ(double(2.0), axis.max());
    EXPECT_EQ(1.5, axis.binCenter(0));
}

TEST(FixedBinAxisTest, FindClosestIndex)
{
    FixedBinAxis v1("name", 2, 0.0, 1.0);
    EXPECT_EQ(size_t(2), v1.size());
    EXPECT_EQ(size_t(0), v1.findClosestIndex(0.0));
    EXPECT_EQ(size_t(0), v1.findClosestIndex(0.25));
    EXPECT_EQ(size_t(1), v1.findClosestIndex(0.5));
    EXPECT_EQ(size_t(1), v1.findClosestIndex(0.6));
    //    EXPECT_THROW( v1.findClosestIndex(1.0), std::runtime_error);
    EXPECT_EQ(size_t(1), v1.findClosestIndex(1.0));

    FixedBinAxis v2("name", 3, -1.5, 1.5);
    EXPECT_EQ(size_t(0), v2.findClosestIndex(-1.5));
    EXPECT_EQ(size_t(0), v2.findClosestIndex(-1.0));
    EXPECT_EQ(size_t(1), v2.findClosestIndex(-0.5));
    EXPECT_EQ(size_t(1), v2.findClosestIndex(0.0));
    EXPECT_EQ(size_t(2), v2.findClosestIndex(0.5));
    EXPECT_EQ(size_t(2), v2.findClosestIndex(1.499));
    //    EXPECT_THROW( v2.findClosestIndex(1.5), std::runtime_error);
    EXPECT_EQ(size_t(2), v2.findClosestIndex(1.5));
}

TEST(FixedBinAxisTest, CheckBin)
{
    FixedBinAxis axis("name", 20, 0, 10);

    Bin1D bin0 = axis.bin(0);
    EXPECT_DOUBLE_EQ(0.25, bin0.center());
    EXPECT_DOUBLE_EQ(0.0, bin0.m_lower);
    EXPECT_DOUBLE_EQ(0.5, bin0.m_upper);
    EXPECT_DOUBLE_EQ(0.5, bin0.binSize());

    Bin1D bin3 = axis.bin(3);
    EXPECT_DOUBLE_EQ(1.75, bin3.center());
    EXPECT_DOUBLE_EQ(1.5, bin3.m_lower);
    EXPECT_DOUBLE_EQ(2.0, bin3.m_upper);
    EXPECT_DOUBLE_EQ(0.5, bin3.binSize());

    Bin1D bin10 = axis.bin(10);
    EXPECT_DOUBLE_EQ(5.25, bin10.center());
    EXPECT_DOUBLE_EQ(5.0, bin10.m_lower);
    EXPECT_DOUBLE_EQ(5.5, bin10.m_upper);

    Bin1D bin19 = axis.bin(19);
    EXPECT_DOUBLE_EQ(9.75, bin19.center());
    EXPECT_DOUBLE_EQ(9.5, bin19.m_lower);
    EXPECT_DOUBLE_EQ(10.0, bin19.m_upper);

    EXPECT_THROW(axis.bin(20), std::runtime_error);

    FixedBinAxis axis2("name", 3, -1, 2.0);
    EXPECT_DOUBLE_EQ(-0.5, axis2.bin(0).center());
    EXPECT_DOUBLE_EQ(0.5, axis2.bin(1).center());
    EXPECT_DOUBLE_EQ(1.5, axis2.bin(2).center());
}

TEST(FixedBinAxisTest, CheckEquality)
{
    FixedBinAxis b1("axis", 99, -1.01, 3.3);
    FixedBinAxis b2("axis", 99, -1.01, 3.3);
    EXPECT_TRUE(b1 == b2);
    FixedBinAxis b3("axissss", 99, -1.01, 3.3);
    FixedBinAxis b4("axis", 99, -1.0, 3.3);
    FixedBinAxis b5("axis", 99, -1.01, 3.29);
    FixedBinAxis b6("axiss", 98, -1.01, 3.3);
    EXPECT_FALSE(b1 == b3);
    EXPECT_FALSE(b1 == b4);
    EXPECT_FALSE(b1 == b5);
    EXPECT_FALSE(b1 == b6);
}

TEST(FixedBinAxisTest, CheckClone)
{
    FixedBinAxis a1("axis", 99, -1.2, 5.4);
    FixedBinAxis* clone = a1.clone();
    EXPECT_TRUE(a1 == *clone);
    delete clone;
}

TEST(FixedBinAxisTest, IOStream)
{
    FixedBinAxis axis("name", 99, -1.2, 5.4);

    std::ostringstream oss;
    oss << axis;
    std::istringstream iss(oss.str());

    std::unique_ptr<IAxis> result(DataUtils::Format::createAxis(iss));
    EXPECT_TRUE(axis == *result);
}

TEST(FixedBinAxisTest, BinCenters)
{
    FixedBinAxis axis("name", 3, -1.5, 1.5);
    std::vector<double> centers = axis.binCenters();
    EXPECT_EQ(size_t(3), centers.size());
    EXPECT_DOUBLE_EQ(-1.0, centers[0]);
    EXPECT_DOUBLE_EQ(0.0, centers[1]);
    EXPECT_DOUBLE_EQ(1.0, centers[2]);

    EXPECT_DOUBLE_EQ(axis.binCenter(0), centers[0]);
    EXPECT_DOUBLE_EQ(axis.binCenter(1), centers[1]);
    EXPECT_DOUBLE_EQ(axis.binCenter(2), centers[2]);
}

TEST(FixedBinAxisTest, BinBoundaries)
{
    FixedBinAxis axis("name", 3, -1.5, 1.5);
    std::vector<double> boundaries = axis.binBoundaries();
    EXPECT_EQ(size_t(4), boundaries.size());
    EXPECT_DOUBLE_EQ(-1.5, boundaries[0]);
    EXPECT_DOUBLE_EQ(-0.5, boundaries[1]);
    EXPECT_DOUBLE_EQ(0.5, boundaries[2]);
    EXPECT_DOUBLE_EQ(1.5, boundaries[3]);
}

TEST(FixedBinAxisTest, ClippedAxis)
{
    FixedBinAxis axis("name", 4, -1.0, 3.0);

    FixedBinAxis* clip1 = axis.clone();
    clip1->clip(-0.5, 2.5);
    EXPECT_EQ(clip1->size(), axis.size());
    EXPECT_EQ(clip1->min(), axis.min());
    EXPECT_EQ(clip1->max(), axis.max());
    EXPECT_TRUE(*clip1 == axis);
    delete clip1;

    FixedBinAxis* clip2 = axis.clone();
    clip2->clip(0.0, 1.99);
    EXPECT_EQ(clip2->size(), size_t(2));
    EXPECT_EQ(clip2->min(), 0.0);
    EXPECT_EQ(clip2->max(), 2.0);
    EXPECT_TRUE(*clip2 != axis);
    delete clip2;
}
