#include "Base/Axis/Bin.h"
#include "Base/Axis/HistoAxis.h"
#include "Device/IO/DataFormatUtils.h"
#include "Tests/GTestWrapper/google_test.h"

TEST(HistoAxisTest, VectorOfUnitLength)
{
    static const double arr[] = {0., 1.};
    std::vector<double> values(arr, arr + sizeof(arr) / sizeof(arr[0]));
    HistoAxis axis("name", values);
    EXPECT_EQ(size_t(1), axis.size());
    EXPECT_EQ(0.0, axis.min());
    EXPECT_EQ(1.0, axis.max());
    EXPECT_EQ(0.5, axis.binCenter(0));
}

TEST(HistoAxisTest, ValidityOfCOnstructor)
{
    static const double arr1[] = {-1.5, -1.5, 0.5, 1.5};
    std::vector<double> v1(arr1, arr1 + sizeof(arr1) / sizeof(arr1[0]));
    EXPECT_THROW(HistoAxis("name", v1), std::runtime_error);

    static const double arr2[] = {1.5, -0.5, 0.5, -1.5};
    std::vector<double> v2(arr2, arr2 + sizeof(arr2) / sizeof(arr2[0]));
    EXPECT_THROW(HistoAxis("name", v2), std::runtime_error);
}

TEST(HistoAxisTest, IndexedAccessor)
{
    std::vector<double> values;

    double start(0.0);
    double end(10.0);
    size_t nbins(100);
    double step = (end - start) / (nbins);
    for (size_t i = 0; i < nbins + 1; ++i)
        values.push_back(start + step * (int)i);

    HistoAxis a1("name", values);

    EXPECT_EQ(100u, a1.size());
    EXPECT_EQ(0.0, a1.min());
    EXPECT_EQ(10.0, a1.max());
    EXPECT_DOUBLE_EQ(0.05, a1.binCenter(0));
    EXPECT_DOUBLE_EQ(0.15, a1.binCenter(1));
    EXPECT_DOUBLE_EQ(6.55, a1.binCenter(65));
    EXPECT_DOUBLE_EQ(9.95, a1.binCenter(99));

    static const double arr[] = {-1.5, -0.5, 0.5, 1.5};
    std::vector<double> v2(arr, arr + sizeof(arr) / sizeof(arr[0]));
    HistoAxis a2("name", v2);
    EXPECT_DOUBLE_EQ(-1.0, a2.binCenter(0));
    EXPECT_DOUBLE_EQ(0.0, a2.binCenter(1));
    EXPECT_DOUBLE_EQ(1.0, a2.binCenter(2));
    EXPECT_THROW(a2.binCenter(3), std::runtime_error);
}

TEST(HistoAxisTest, FindClosestIndex)
{
    static const double arr1[] = {0.0, 0.5, 1.0};
    std::vector<double> values1(arr1, arr1 + sizeof(arr1) / sizeof(arr1[0]));

    HistoAxis v1("name", values1);
    EXPECT_EQ(size_t(2), v1.size());
    EXPECT_EQ(size_t(0), v1.findClosestIndex(0.0));
    EXPECT_EQ(size_t(0), v1.findClosestIndex(0.25));
    EXPECT_EQ(size_t(1), v1.findClosestIndex(0.5));
    EXPECT_EQ(size_t(1), v1.findClosestIndex(0.6));
    //    EXPECT_THROW( v1.findClosestIndex(1.0), std::runtime_error);
    EXPECT_EQ(size_t(1), v1.findClosestIndex(1.0));

    static const double arr2[] = {-1.5, -0.5, 0.5, 1.5};
    std::vector<double> values2(arr2, arr2 + sizeof(arr2) / sizeof(arr2[0]));
    HistoAxis v2("name", values2);
    EXPECT_EQ(size_t(0), v2.findClosestIndex(-1.5));
    EXPECT_EQ(size_t(0), v2.findClosestIndex(-1.0));
    EXPECT_EQ(size_t(1), v2.findClosestIndex(-0.5));
    EXPECT_EQ(size_t(1), v2.findClosestIndex(0.0));
    EXPECT_EQ(size_t(2), v2.findClosestIndex(0.5));
    EXPECT_EQ(size_t(2), v2.findClosestIndex(1.499));
    //    EXPECT_THROW( v2.findClosestIndex(1.5), std::runtime_error);
    EXPECT_EQ(size_t(2), v2.findClosestIndex(1.5));

    static const double arr3[] = {-1.0, -0.5, 0.5, 1.0, 2.0};
    std::vector<double> values3(arr3, arr3 + sizeof(arr3) / sizeof(arr2[0]));
    HistoAxis v3("name", values3);
    EXPECT_EQ(size_t(0), v3.findClosestIndex(-1.0));
    EXPECT_EQ(size_t(0), v3.findClosestIndex(-0.75));
    EXPECT_EQ(size_t(0), v3.findClosestIndex(-0.51));
    EXPECT_EQ(size_t(1), v3.findClosestIndex(-0.5));
    EXPECT_EQ(size_t(1), v3.findClosestIndex(0.1));
    EXPECT_EQ(size_t(1), v3.findClosestIndex(0.4999));
    EXPECT_EQ(size_t(2), v3.findClosestIndex(0.5));
    EXPECT_EQ(size_t(2), v3.findClosestIndex(0.99999));
    EXPECT_EQ(size_t(3), v3.findClosestIndex(1.0));
    EXPECT_EQ(size_t(3), v3.findClosestIndex(1.9999));
}

TEST(HistoAxisTest, CheckBin)
{
    static const double arr3[] = {-1.0, -0.5, 0.5, 1.0, 2.0};
    std::vector<double> values3(arr3, arr3 + sizeof(arr3) / sizeof(arr3[0]));
    HistoAxis axis("name", values3);
    EXPECT_DOUBLE_EQ(-1.0, axis.bin(0).m_lower);
    EXPECT_DOUBLE_EQ(-0.75, axis.bin(0).center());
    EXPECT_DOUBLE_EQ(-0.5, axis.bin(0).m_upper);
    EXPECT_DOUBLE_EQ(0.5, axis.bin(0).binSize());

    EXPECT_DOUBLE_EQ(-0.5, axis.bin(1).m_lower);
    EXPECT_DOUBLE_EQ(0.0, axis.bin(1).center());
    EXPECT_DOUBLE_EQ(0.5, axis.bin(1).m_upper);
    EXPECT_DOUBLE_EQ(1.0, axis.bin(1).binSize());

    EXPECT_DOUBLE_EQ(0.5, axis.bin(2).m_lower);
    EXPECT_DOUBLE_EQ(0.75, axis.bin(2).center());
    EXPECT_DOUBLE_EQ(1.0, axis.bin(2).m_upper);
    EXPECT_DOUBLE_EQ(0.5, axis.bin(2).binSize());

    EXPECT_DOUBLE_EQ(1.0, axis.bin(3).m_lower);
    EXPECT_DOUBLE_EQ(1.5, axis.bin(3).center());
    EXPECT_DOUBLE_EQ(2.0, axis.bin(3).m_upper);
    EXPECT_DOUBLE_EQ(1.0, axis.bin(3).binSize());
}

TEST(HistoAxisTest, CheckEquality)
{
    static const double arr3[] = {-1.0, -0.5, 0.5, 1.0, 2.0};
    std::vector<double> values3(arr3, arr3 + sizeof(arr3) / sizeof(arr3[0]));

    HistoAxis a1("name", values3);
    HistoAxis a2("name", values3);
    HistoAxis a3("nameX", values3);
    EXPECT_TRUE(a1 == a2);
    EXPECT_FALSE(a1 == a3);

    static const double arr4[] = {-1.0, -0.5, 0.5, 1.1, 2.0};
    std::vector<double> values4(arr4, arr4 + sizeof(arr4) / sizeof(arr4[0]));
    HistoAxis a4("nameX", values4);
    EXPECT_FALSE(a1 == a4);
}

TEST(HistoAxisTest, CheckClone)
{
    static const double arr3[] = {-1.0, -0.5, 0.5, 1.0, 2.0};
    std::vector<double> values3(arr3, arr3 + sizeof(arr3) / sizeof(arr3[0]));
    HistoAxis a1("name", values3);

    HistoAxis* clone = a1.clone();
    EXPECT_TRUE(a1 == *clone);
    delete clone;
}

TEST(HistoAxisTest, IOStream)
{
    static const double arr[] = {-1.0, -0.5, 0.5, 1.0, 2.0};
    std::vector<double> values(arr, arr + sizeof(arr) / sizeof(arr[0]));
    HistoAxis axis("name", values);

    std::ostringstream oss;
    oss << axis;
    std::istringstream iss(oss.str());

    std::unique_ptr<IAxis> result(DataUtils::Format::createAxis(iss));
    EXPECT_TRUE(axis == *result);
}

TEST(HistoAxisTest, BinCenters)
{
    static const double arr[] = {-1.0, -0.5, 0.5, 1.0, 2.0};
    std::vector<double> values(arr, arr + sizeof(arr) / sizeof(arr[0]));
    HistoAxis axis("name", values);

    std::vector<double> centers = axis.binCenters();
    EXPECT_EQ(size_t(4), centers.size());
    EXPECT_DOUBLE_EQ(-0.75, centers[0]);
    EXPECT_DOUBLE_EQ(0.0, centers[1]);
    EXPECT_DOUBLE_EQ(0.75, centers[2]);
    EXPECT_DOUBLE_EQ(1.5, centers[3]);
}

TEST(HistoAxisTest, BinBoundaries)
{
    static const double arr[] = {-1.0, -0.5, 0.5, 1.0, 2.0};
    std::vector<double> values(arr, arr + sizeof(arr) / sizeof(arr[0]));
    HistoAxis axis("name", values);

    std::vector<double> boundaries = axis.binBoundaries();
    EXPECT_EQ(size_t(5), boundaries.size());
    EXPECT_DOUBLE_EQ(-1.0, boundaries[0]);
    EXPECT_DOUBLE_EQ(-0.5, boundaries[1]);
    EXPECT_DOUBLE_EQ(0.5, boundaries[2]);
    EXPECT_DOUBLE_EQ(1.0, boundaries[3]);
    EXPECT_DOUBLE_EQ(2.0, boundaries[4]);
}

TEST(HistoAxisTest, ClippedAxis)
{
    static const double arr[] = {-1.0, -0.5, 0.5, 1.0, 2.0};
    std::vector<double> values(arr, arr + sizeof(arr) / sizeof(arr[0]));
    HistoAxis axis("name", values);

    HistoAxis* clip1 = axis.clone();
    clip1->clip(-1.0, 2.0);
    EXPECT_TRUE(axis == *clip1);
    delete clip1;

    HistoAxis* clip2 = axis.clone();
    clip2->clip(-0.5, 1.5);
    EXPECT_EQ(clip2->size(), size_t(3));
    EXPECT_EQ(clip2->min(), -0.5);
    EXPECT_EQ(clip2->max(), 2.0);
    std::vector<double> centers = clip2->binCenters();
    EXPECT_EQ(centers[0], 0.0);
    EXPECT_EQ(centers[1], 0.75);
    EXPECT_EQ(centers[2], 1.5);
    EXPECT_TRUE(axis != *clip2);
    delete clip2;

    HistoAxis* clip3 = axis.clone();
    clip3->clip(-0.5, 0.99);
    EXPECT_EQ(clip3->size(), size_t(2));
    EXPECT_EQ(clip3->min(), -0.5);
    EXPECT_EQ(clip3->max(), 1.0);
    std::vector<double> boundaries = clip3->binBoundaries();
    EXPECT_EQ(boundaries[0], -0.5);
    EXPECT_EQ(boundaries[1], 0.5);
    EXPECT_EQ(boundaries[2], 1.0);
    delete clip3;
}
